@@ -51,22 +51,6 @@ void sigalrm_handler( int signo )
    sigalrm = 1;
 }
 
-#define ATOM_SYMBOL     1
-#define ATOM_STRING     2
-#define ATOM_FIXNUM     3
-#define ATOM_REGEXP     4
-#define ATOM_TABLE      5
-#define ATOM_STACK      6
-#define ATOM_INTRINSIC  7
-#define ATOM_DB         8
-#define ATOM_CLOSURE    9
-#define ATOM_MACRO      10
-#define ATOM_ACT_RECORD 11
-#define ATOM_RECORD     12
-#define ATOM_SQL        13
-#define ATOM_LIST       14
-#define ATOM_ATOM       15
-
 void print_err( int err, char *func, int arg, int type )
 {
    static char *types[] = { "", "symbol", "string", "fixnum",
@@ -93,9 +77,9 @@ void *memory( int size )
    if ( size == 0 )
       return NULL;
 
-   if (( ptr = malloc( size )) == NULL )
+   if (( ptr = calloc( 1, size )) == NULL )
    {
-      fprintf( stderr, "memory: malloc: %s.\n", strerror( errno ));
+      fprintf( stderr, "memory: calloc: %s.\n", strerror( errno ));
       exit( 1 );
    }
 
@@ -185,6 +169,13 @@ unsigned int hash_key( char *name, unsigned int len )
    unsigned int i, key;
    char *nptr;
 
+   if ( len > INT_MAX )
+   {
+      fprintf( stderr, "objects cannot have print syntax longer than "
+                       "(maxidx): %d.\n", INT_MAX );
+      exit( 1 );
+   }
+
    for( i = 0, key = 2166136261, nptr = name; i < len; ++nptr, ++i )
       key = ( key * 16777619 ) ^ *nptr;
 
@@ -193,12 +184,12 @@ unsigned int hash_key( char *name, unsigned int len )
 
 int insert_elt( struct hash_elt **hash, struct atom *atom, struct object *object, int size )
 {
-   unsigned int key;
+   unsigned int orig, key;
    struct hash_elt *ptr = NULL, *ptr2 = NULL;
    int found = 0;
-      
-   key = hash_key( atom->syntax, atom->len );
-   key %= size;
+
+   orig = hash_key( atom->syntax, atom->len );
+   key = orig % size;
 
    if ( hash[ key ] == NULL )
    {
@@ -206,6 +197,7 @@ int insert_elt( struct hash_elt **hash, struct atom *atom, struct object *object
       hash[ key ]->element.ptr = object;
       hash[ key ]->next = NULL;
       hash[ key ]->key = atom;
+      hash[ key ]->hash_key = orig;
       return found;
    }
 
@@ -225,6 +217,7 @@ int insert_elt( struct hash_elt **hash, struct atom *atom, struct object *object
       ptr2->next = make_hash_elt();
       ptr = ptr2->next;
       ptr->next = NULL;
+      ptr->hash_key = orig;
    }
 
    ptr->element.ptr = object;
@@ -235,11 +228,11 @@ int insert_elt( struct hash_elt **hash, struct atom *atom, struct object *object
 
 void insert_elt_int( struct hash_elt **hash, struct atom *atom, int elt )
 {
-   unsigned int key;
+   unsigned int orig, key;
    struct hash_elt *ptr = NULL, *ptr2 = NULL;
-      
-   key = hash_key( atom->syntax, atom->len );
-   key %= HASH_SIZE;
+
+   orig = hash_key( atom->syntax, atom->len );
+   key = orig % HASH_SIZE;
 
    if ( hash[ key ] == NULL )
    {
@@ -247,6 +240,7 @@ void insert_elt_int( struct hash_elt **hash, struct atom *atom, int elt )
       hash[ key ]->element.integer = elt;
       hash[ key ]->next = NULL;
       hash[ key ]->key = atom;
+      hash[ key ]->hash_key = orig;
       return;
    }
 
@@ -263,6 +257,7 @@ void insert_elt_int( struct hash_elt **hash, struct atom *atom, int elt )
       ptr2->next = make_hash_elt();
       ptr = ptr2->next;
       ptr->next = NULL;
+      ptr->hash_key = orig;
    }
 
    ptr->element.integer = elt;
@@ -273,7 +268,7 @@ void *lookup_elt( struct hash_elt **hash, struct atom *atom, int size )
 {
    unsigned int key;
    struct hash_elt *ptr;
-      
+
    key = hash_key( atom->syntax, atom->len );
    key %= size;
 
@@ -298,7 +293,7 @@ int lookup_elt_int( struct hash_elt **hash, struct atom *atom )
 {
    unsigned int key;
    struct hash_elt *ptr;
-      
+
    key = hash_key( atom->syntax, atom->len );
    key %= HASH_SIZE;
 
@@ -323,7 +318,7 @@ int remove_elt( struct hash_elt **hash, struct atom *atom, int size )
 {
    unsigned int key;
    struct hash_elt *ptr, *ptr2;
-      
+
    key = hash_key( atom->syntax, atom->len );
    key %= size;
 
@@ -357,17 +352,12 @@ int remove_elt( struct hash_elt **hash, struct atom *atom, int size )
    return 1;
 }
 
-void remove_id( char *name, int len, int id )
+void remove_id( char *name, unsigned int len, int id )
 {
    unsigned int key;
-   char *nptr;
    struct hash_elt *ptr, *ptr2;
 
-   --syntax_items;
-
-   for( key = 2166136261, nptr = name; *nptr; ++nptr )
-      key = ( key * 16777619 ) ^ *nptr;
-
+   key = hash_key( name, len );
    key %= syntax_size;
 
    if ( syntax[ key ] == NULL )
@@ -389,12 +379,14 @@ void remove_id( char *name, int len, int id )
    if ( ptr == NULL )
       return;
 
+   --syntax_items;
+
    if ( ptr2 == NULL )
       syntax[ key ] = ptr->next;
    else
       ptr2->next = ptr->next;
 
-   free_atom( ( struct atom *)ptr->element.ptr );
+   free_atom(( struct atom *)ptr->element.ptr );
 
    bzero( ptr, sizeof( struct hash_elt ));
    STACK_PUSH( reclaimed_hash_elts, ptr )
@@ -403,8 +395,8 @@ void remove_id( char *name, int len, int id )
 void resize_syntax()
 {
    struct hash_elt **new, **old, **ptr, *ptr2, *ptr3, *nptr, *nptr2;
-   char *sptr;
-   int size, i, n, old_size, idx;
+   int size, i, old_size;
+   unsigned int idx;
 
    if ( syntax_size * HASH_RESIZE < syntax_size )
       return;
@@ -416,8 +408,6 @@ void resize_syntax()
 #endif
 
    new = memory( size * sizeof( struct hash_elt * ));
-   bzero( new, size * sizeof( struct hash_elt * ));
-
    old = syntax;
    old_size = syntax_size;
 
@@ -434,10 +424,8 @@ void resize_syntax()
       {
          ptr3 = ptr2->next;
 
-         for( idx = 0, n = ptr2->key->len, sptr = ptr2->key->syntax; n; --n, ++sptr )
-            idx = ( idx << 4 ) ^ ( idx >> 28 ) ^ *sptr;
-
-         idx = abs( idx ) % size;
+         idx = ptr2->hash_key;
+         idx %= size;
 
          if ( syntax[ idx ] == NULL )
          {
@@ -465,15 +453,15 @@ void resize_syntax()
 
 struct atom *get_id( char *name, int unsigned len, int duplicate )
 {
-   unsigned int key;
+   unsigned int orig, key;
    struct atom *atom;
    struct hash_elt *ptr = NULL, *ptr2 = NULL;
 
    if ( syntax_items > syntax_resize )
       resize_syntax();
 
-   key = hash_key( name, len );
-   key %= syntax_size;
+   orig = hash_key( name, len );
+   key = orig % syntax_size;
 
    if ( syntax[ key ] == NULL )
    {
@@ -495,6 +483,8 @@ struct atom *get_id( char *name, int unsigned len, int duplicate )
       syntax[ key ]->element.ptr = ( struct object *)atom;
       syntax[ key ]->next = NULL;
       syntax[ key ]->key = atom;
+      syntax[ key ]->hash_key = orig;
+      ++syntax_items;
 
       return atom;
    }
@@ -503,30 +493,29 @@ struct atom *get_id( char *name, int unsigned len, int duplicate )
 
    for( ptr = ptr2; ptr != NULL; ptr = ptr->next )
    {
+      char *ptr3, *ptr4;
+      unsigned int len1, len2;
+
       atom = ( struct atom *)ptr->element.ptr;
 
-      if ( len == atom->len )
-      {
-         char *ptr3 = name, *ptr4 = atom->syntax;
-         int idx = len;
+      for( ptr3 = name, ptr4 = atom->syntax, len1 = len, len2 = atom->len;
+           len1 && len2;
+           ++ptr3, ++ptr4, --len1, --len2 )
 
-         while( idx-- )
-            if ( *ptr3++ != *ptr4++ )
-               break;
+         if ( *ptr3 != *ptr4 )
+            break;
 
-         if ( idx < 0 )
-            return atom;
-      }
+      if ( ! len1 && ! len2 )
+         return atom;
 
       ptr2 = ptr;
    }
 
-   if ( ptr == NULL )
-   {
-      ptr2->next = make_hash_elt();
-      ptr = ptr2->next;
-      ptr->next = NULL;
-   }
+   ptr2->next = make_hash_elt();
+   ptr = ptr2->next;
+   ptr->next = NULL;
+   ptr->hash_key = orig;
+   ++syntax_items;
 
    atom = make_atom();
    atom->syntax = ( duplicate ? str_dup( name, len ) : name );
@@ -924,10 +913,6 @@ void initialize( int argc, char **argv )
    hash_elt_pool_free = POOL_INC;
    hash_elt_pool_ptr = hash_elt_pool;
 
-   bzero( object_pool, POOL_INC );
-   bzero( atom_pool, POOL_INC );
-   bzero( hash_elt_pool, POOL_INC );
-
    stack = make_stack();
    open_envs = make_stack();
 
@@ -1404,7 +1389,7 @@ int get_token( int depth, int file )
 
          if ( type == TOK_STRING && escape && *ptr != '\\' && *ptr != '"' )
             escape = 0;
-         
+
          if ( *ptr == '\\' )
          {
             if ( type != TOK_STRING )
@@ -1412,7 +1397,7 @@ int get_token( int depth, int file )
                ++ptr;
                continue;
             }
-            
+
             if ( !( escape ^= 1 ) )
                STRING_CHOP( token )
          }
@@ -1457,7 +1442,7 @@ int get_token( int depth, int file )
 
             if ( type != TOK_STRING )
                return type;
-            
+
             if ( escape )
                STRING_CHOP( token )
 
@@ -1647,7 +1632,7 @@ int process_token( int type, int depth )
 
       ptr->data.atom = entry;
    }
-      
+
    STACK_PUSH(( depth ? input_stack : stack ), ptr )
 
    if ( quote == -1 )
@@ -2821,6 +2806,10 @@ void sweep_objects()
    struct object *top;
    int i, j, old_stack_inc;
 
+   reclaimed_objects->free += reclaimed_objects->used;
+   reclaimed_objects->used = 0;
+   reclaimed_objects->top = reclaimed_objects->values;
+   
    old_stack_inc = stack_inc;
    stack_inc = 65536;
 
@@ -2831,14 +2820,17 @@ void sweep_objects()
       top = ( struct object *)object_pool_stack->values[ --i ].ptr;
 
       for( j = 0; j < POOL_INC; ++j, ++top )
+      {
          if ( ! ismarked( top->flags ) && ( top->flags || top->data.atom != NULL ))
          {
             --objects_allocated;
             bzero( top, sizeof( struct object ));
             STACK_PUSH( reclaimed_objects, top )
+            continue;
          }
-         else
-            unmark( top->flags );
+
+         unmark( top->flags );
+      }
    }
 
    for( top = object_pool; top < object_pool_ptr; ++top )
@@ -2848,9 +2840,10 @@ void sweep_objects()
          --objects_allocated;
          bzero( top, sizeof( struct object ));
          STACK_PUSH( reclaimed_objects, top )
+         continue;
       }
-      else
-         unmark( top->flags );
+
+      unmark( top->flags );
    }
 
    stack_inc = old_stack_inc;
@@ -2915,6 +2908,10 @@ void sweep_atoms()
    struct atom *top;
    int i, j, old_stack_inc;
 
+   reclaimed_atoms->free += reclaimed_atoms->used;
+   reclaimed_atoms->used = 0;
+   reclaimed_atoms->top = reclaimed_atoms->values;
+   
    old_stack_inc = stack_inc;
    stack_inc = 65536;
 
@@ -2926,10 +2923,12 @@ void sweep_atoms()
 
       for( j = 0; j < POOL_INC; ++j, ++top )
       {
-         if ( !top->id || top->id == lambda_id || top->id == macro_id || top->id == quote_id ||
-              top->id == underscore_id || type( top->flags ) == ATOM_INTRINSIC )
+         if ( !top->id || top->id == lambda_id || top->id == macro_id || 
+              top->id == quote_id || top->id == underscore_id || 
+              type( top->flags ) == ATOM_INTRINSIC )
             continue;
-         else if ( top->id && ! isitset( top->flags ))
+
+         if ( top->id && ! isitset( top->flags ))
          {
             --atoms_allocated;
 
@@ -2938,18 +2937,21 @@ void sweep_atoms()
 
             bzero( top, sizeof( struct atom ));
             STACK_PUSH( reclaimed_atoms, top )
+            continue;
          }
-         else
-            unset( top->flags );
+
+         unset( top->flags );
       }
    }
 
    for( top = atom_pool; top < atom_pool_ptr; ++top )
    {
-      if ( !top->id || top->id == lambda_id || top->id == macro_id || top->id == quote_id ||
-           top->id == underscore_id || type( top->flags ) == ATOM_INTRINSIC )
+      if ( !top->id || top->id == lambda_id || top->id == macro_id || 
+           top->id == quote_id || top->id == underscore_id || 
+           type( top->flags ) == ATOM_INTRINSIC )
          continue;
-      else if ( top->id && ! isitset( top->flags ))
+
+      if ( top->id && ! isitset( top->flags ))
       {
          --atoms_allocated;
 
@@ -2958,9 +2960,10 @@ void sweep_atoms()
 
          bzero( top, sizeof( struct atom ));
          STACK_PUSH( reclaimed_atoms, top )
+         continue;
       }
-      else
-         unset( top->flags );
+
+      unset( top->flags );
    }
 
    stack_inc = old_stack_inc;
@@ -2995,9 +2998,6 @@ void mark_bookmarks()
 void collect_garbage()
 {
 #ifdef DEBUG
-   int old_objects = reclaimed_objects->used;
-   int old_atoms = reclaimed_atoms->used;
-
    fputs( "[collecting garbage: marking...", stderr );
 #endif
 
@@ -3023,8 +3023,7 @@ void collect_garbage()
 
 #ifdef DEBUG
    fprintf( stderr, "done: objects reclaimed: %d, atoms reclaimed: %d]\n",
-            reclaimed_objects->used - old_objects,
-            reclaimed_atoms->used - old_atoms );
+            reclaimed_objects->used, reclaimed_atoms->used );
 #endif
 }
 
